@@ -1,12 +1,21 @@
+import logging
+import base64
+import os
+import warnings
+
 import copy as copy_lib
 
+from pathlib import Path
 from dataclasses import dataclass, field
 from typing import Any
-import logging
+
+from bs4 import BeautifulSoup
 
 from .html import Body, Container, Row, Column
-from .utils import as_list, get_filename, get_valid_list_idx, limit_list_insert_idx
-from .recipes import get_html
+from .utils import as_list, get_filename, get_valid_list_idx, limit_list_insert_idx, check_is_url, get_image_data, Working_Directory
+from .recipes import get_html, get_recipes, get_recipe_info
+from .rendering import get_renderers, get_renderer_info
+
 
 @dataclass
 class Report:
@@ -15,13 +24,14 @@ class Report:
     """
     name : str = "BakePy_Report"
 
-    main_stylesheet: str = """<link href="https://cdn.jsdelivr.net/npm/bootstrap@5.2.1/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-iYQeCzEYFbKjA/T2uDLTpkwGzCiq6soy8tYaI1GyVh/UjpbCx/TYkiZhlZB6+fzT" crossorigin="anonymous">"""
+    main_stylesheet: str = """<link href="https://cdn.jsdelivr.net/npm/bootstrap@5.2.1/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-iYQeCzEYFbKjA/T2uDLTpkwGzCiq6soy8tYaI1GyVh/UjpbCx/TYkiZhlZB6+fzT" crossorigin="anonymous">
+                              <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/katex@0.16.2/dist/katex.min.css" integrity="sha384-bYdxxUwYipFNohQlHt0bjN/LCpueqWz13HufFEV1SUatKs1cm4L6fFgCi1jT643X" crossorigin="anonymous">"""
     stylesheets: Any = field(default_factory=list)
 
     cont_default_styles: Any = field(default_factory=list)
-    cont_default_classes: Any = field(default_factory=list)
+    cont_default_classes: Any = field(default_factory=lambda:["my-5"])
     row_default_styles: Any = field(default_factory=list)
-    row_default_classes: Any = field(default_factory=lambda:["justify-content-center pb-5 gx-5"])
+    row_default_classes: Any = field(default_factory=lambda:["justify-content-center mb-5 gx-5"])
     col_default_styles: Any = field(default_factory=list)
     col_default_classes: Any = field(default_factory=list)
 
@@ -146,7 +156,7 @@ class Report:
                 self.current_container = self.containers[self.current_container_name]
                 logging.info(f"Set current container to {self.current_container_name}.")
         #Verify the current row
-        if self.current_row is None or self.current_row not in self.current_container.elements:
+        if self.current_row is None or not any([self.current_row is x for x in self.current_container.elements]):
             self.current_col = None
             self.current_col_idx = None
             c_items = len(self.current_container.elements)
@@ -160,7 +170,7 @@ class Report:
                 self.current_row = self.current_container.elements[self.current_row_idx]
                 logging.info(f"Set current row to {self.current_row_idx}.")
         #Verify the current column
-        if self.current_col is None or self.current_col not in self.current_row.elements:
+        if self.current_col is None or not any([self.current_col is x for x in self.current_row.elements]):
             c_items = len(self.current_row.elements)
             if c_items == 0:
                 self.current_col = None
@@ -596,7 +606,7 @@ class Report:
         overwrite : bool, default = False
             Set to true if overwriting the current classes.
         """
-        e = self._get_row_from_idx_arg(idx = row_idx, container_name = container_name)
+        e = self._get_row_from_idx_arg(row_idx = row_idx, container_name = container_name)
         e.add_cls(vals, overwrite)
     
     def set_row_sty(self, vals, row_idx = None, container_name = None, overwrite = False):
@@ -614,7 +624,7 @@ class Report:
         overwrite : bool, default = False
             Set to true if overwriting the current styles.
         """
-        e = self._get_row_from_idx_arg(idx = row_idx, container_name = container_name)
+        e = self._get_row_from_idx_arg(row_idx = row_idx, container_name = container_name)
         e.add_sty(vals, overwrite)
 
     def set_col_cls(self, vals, col_idx = None, row_idx = None, container_name = None, overwrite = False):
@@ -662,7 +672,7 @@ class Report:
             container_name = None,
             row_idx = None,
             col_idx = None,
-            new_row = False,
+            new_row = True,
             new_col = True,
             size = None,
             copy = False,
@@ -681,20 +691,20 @@ class Report:
             The index of the row to insert at. If None, uses the current row.
         col_idx : int, default = None
             The index of the column to insert at. If None, inserts at the end of the row.
-        new_row : bool, default = False
+        new_row : bool, default = True
             If True, the elements will be added in a new row.
         new_col : bool, default = True
-            If True, the elements will be added in a new column.
+            If True, the elements will be added in a new column. Takes precedence over new_row.
         size : int, default = None
             The width of the column to insert.
         copy : bool, default = False
             If True, the elements will be copied before inserting.
         overwrite : bool, default = False
-            Set to true if overwriting the current styles.
+            Set to true if overwriting the item in the specified position.
         """
 
         elements = as_list(elements)
-        if new_row and not overwrite:
+        if new_col and new_row and not overwrite:
             self.add_row(row_idx = row_idx, container_name = container_name, overwrite = False)
         if new_col and not overwrite:
             self.add_col(col_idx = col_idx, row_idx = row_idx, container_name = container_name, overwrite = False)
@@ -714,18 +724,18 @@ class Report:
             else:
                 col.add_element(e, **other_args)
         
-    def add_special(self,
-                    type,
-                    *args,
-                    container_name = None,
-                    row_idx = None,
-                    col_idx = None,
-                    new_row = False,
-                    new_col = True,
-                    size = None,
-                    copy = False,
-                    overwrite = False,
-                    **kwargs):
+    def recipe(self,
+               type,
+               *args,
+               container_name = None,
+               row_idx = None,
+               col_idx = None,
+               new_row = True,
+               new_col = True,
+               size = None,
+               copy = False,
+               overwrite = False,
+               **kwargs):
         """
         Add specially formatted content to the report.
 
@@ -739,7 +749,7 @@ class Report:
             The index of the row to insert at. If None, uses the current row.
         col_idx : int, default = None
             The index of the column to insert at. If None, inserts at the end of the row.
-        new_row : bool, default = False
+        new_row : bool, default = True
             If True, the elements will be added in a new row.
         new_col : bool, default = True
             If True, the elements will be added in a new column.
@@ -748,12 +758,52 @@ class Report:
         copy : bool, default = False
             If True, the elements will be copied before inserting.
         overwrite : bool, default = False
-            Set to true if overwriting the current styles.
+            Set to true if overwriting the item in the specified position.
         """
 
         self.add(get_html(type, *args, **kwargs), container_name, row_idx, col_idx, new_row, new_col, size, copy, overwrite)
-        
-    def save_html(self, filename = None):
+
+    @staticmethod
+    def get_recipes():
+        """
+        Get all available recipes.
+        """
+        return get_recipes()
+
+    @staticmethod
+    def get_recipe_info(recipe):
+        """
+        Get information about a recipe.
+
+        Parameters
+        ----------
+        recipe: str
+            The name of a recipe.
+        """
+        return get_recipe_info(recipe)
+
+    @staticmethod
+    def get_renderers():
+        """
+        Get all available renderers.
+        """
+        return get_renderers()
+
+    @staticmethod
+    def get_renderer_info(recipe, verbose = False):
+        """
+        Get information about the renderer associated to an element.
+
+        Parameters
+        ----------
+        element: object
+            An object.
+        verbose: bool, default = False
+            If True, prints the process of finding the renderer.
+        """
+        return get_renderer_info(recipe, verbose)
+
+    def save_html(self, filename = None, embed_images=True, embed_links=True):
         """
         Save the report to an HTML file.
 
@@ -761,28 +811,67 @@ class Report:
         ----------
         filename : str, default = None
             The path to save at. If None, uses the report name.
+        embed_images : bool, default = True
+            If True, embeds any linked images as base64.
+        embed_links : bool, default = True
+            If True, embeds any linked styles.
         """
-        self.body.save_html(filename)
 
-    def save_pdf(self, filename = None):
-        """
-        Save the report to an PDF file.
-
-        Parameters
-        ----------
-        filename : str, default = None
-            The path to save at. If None, uses the report name.
-        """
-        import weasyprint
-
-        name = self.name
-        
-        if name == "":
-            name = id(self)
-            
         if filename is None:
-            filename = name
+            output_dir = Path().cwd().absolute()
+        else:
+            output_dir = Path(filename).parent.absolute()
+
+        with Working_Directory(output_dir) as wd:
+
+            html = self.body.to_html()
+            soup = BeautifulSoup(html, "html.parser")
+
+            #Embed images as base64
+            if embed_images:
+                for img in soup.find_all('img'):
+                    img_src = img.attrs['src']
+                    #If image is already embedded, skip.
+                    if img_src.startswith('data:'):
+                        continue
+                    img_type, img_data = get_image_data(img_src)                   
+                    img.attrs['src'] = f"data:{img_type};base64,{str(base64.b64encode(img_data),'utf-8')}"
+            #Embed links (stylesheets) using bs4
+            #TODO
+
+            #Save the file
+            name = self.name
             
-        filename = get_filename(filename, "pdf")
+            if name == "":
+                name = id(self)
+                
+            if filename is None:
+                filename = name
+                
+            filename = get_filename(filename, "html")
+
+            with open(filename, 'w') as f:
+                f.write(soup.prettify())
+
+    # def save_pdf(self, filename = None):
+    #     """
+    #     Save the report to an PDF file.
+
+    #     Parameters
+    #     ----------
+    #     filename : str, default = None
+    #         The path to save at. If None, uses the report name.
+    #     """
+    #     import weasyprint
+
+    #     name = self.name
+        
+    #     if name == "":
+    #         name = id(self)
             
-        weasyprint.HTML(string=self.save_html()).write_pdf(filename)
+    #     if filename is None:
+    #         filename = name
+            
+    #     filename = get_filename(filename, "pdf")
+            
+    #     weasyprint.HTML(string=self.save_html(embed_images=True, embed_links=True)).write_pdf(filename)
